package ru.dbotthepony.mc.otm.matter

import net.minecraft.data.CachedOutput
import net.minecraft.data.DataGenerator
import net.minecraft.data.DataProvider
import net.minecraft.resources.ResourceLocation
import net.minecraft.tags.TagKey
import net.minecraft.world.item.Item
import net.minecraft.world.level.ItemLike
import net.minecraftforge.data.event.GatherDataEvent
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.registryName
import java.util.Collections
import java.util.function.Consumer

open class MatterDataProvider(protected val dataGenerator: DataGenerator, val namespace: String?) : DataProvider {
	constructor(event: GatherDataEvent) : this(event.generator, event.modContainer.namespace)

	protected val pathProvider: DataGenerator.PathProvider = dataGenerator.createPathProvider(DataGenerator.Target.DATA_PACK, MatterManager.DIRECTORY)
	protected val actions = LinkedHashMap<ResourceLocation, MatterManager.AbstractRegistryAction>()

	sealed class Configuration(val name: ResourceLocation) {
		var errorOnFailure: Boolean = false
		var matter: ImpreciseFraction? = null
		var complexity: Double? = null
		var priority: Int? = null

		private var _key: ResourceLocation? = null
		private var _tag: TagKey<Item>? = null

		var key: ResourceLocation?
			get() = _key
			set(value) {
				checkNotNull(value) { "Can't set key to null" }
				_tag = null
				_key = value
			}

		var tag: TagKey<Item>?
			get() = _tag
			set(value) {
				checkNotNull(value) { "Can't set tag to null" }
				_key = null
				_tag = value
			}

		protected fun checkKeyTag() {
			check(_key != null || _tag != null) { "You must define either key or tag for $name" }
		}

		protected fun checkMatterValues() {
			check(matter == null || (matter ?: throw ConcurrentModificationException()) >= ImpreciseFraction.ZERO) { "Invalid matter value: ${matter ?: throw ConcurrentModificationException()}" }
			check(complexity == null || (complexity ?: throw ConcurrentModificationException()) >= 0.0) { "Invalid matter value: ${matter ?: throw ConcurrentModificationException()}" }
		}

		abstract fun build(): MatterManager.AbstractRegistryAction
	}

	class InsertConfiguration(name: ResourceLocation) : Configuration(name) {
		var replaceIfExists: Boolean = false
		var comparePriority: Boolean = false

		override fun build(): MatterManager.InsertAction {
			checkKeyTag()
			checkMatterValues()
			check(matter != null && complexity != null) { "You must define both matter value and complexity for $name" }

			if (key != null) {
				return MatterManager.InsertAction(
					key = key ?: throw ConcurrentModificationException(),
					matter = matter,
					complexity = complexity,
					priority = priority,
					errorOnFailure = errorOnFailure,
					replaceIfExists = replaceIfExists,
					comparePriority = comparePriority
				)
			} else {
				return MatterManager.InsertAction(
					tag = tag ?: throw ConcurrentModificationException(),
					matter = matter,
					complexity = complexity,
					priority = priority,
					errorOnFailure = errorOnFailure,
					replaceIfExists = replaceIfExists,
					comparePriority = comparePriority
				)
			}
		}
	}

	class DeleteConfiguration(name: ResourceLocation) : Configuration(name) {
		override fun build(): MatterManager.DeleteAction {
			checkKeyTag()

			check(matter == null) { "Delete action can't have matter value" }
			check(complexity == null) { "Delete action can't have complexity value" }

			if (key != null) {
				return MatterManager.DeleteAction(
					key = key ?: throw ConcurrentModificationException(),
					errorOnFailure = errorOnFailure,
				)
			} else {
				return MatterManager.DeleteAction(
					tag = tag ?: throw ConcurrentModificationException(),
					errorOnFailure = errorOnFailure,
				)
			}
		}
	}

	class UpdateConfiguration(name: ResourceLocation) : Configuration(name) {
		val matterFunctions = ArrayList<MatterManager.UpdateAction.BoundFunction<ImpreciseFraction>>()
		val complexityFunctions = ArrayList<MatterManager.UpdateAction.BoundFunction<Double>>()
		val priorityFunctions = ArrayList<MatterManager.UpdateAction.BoundFunction<Int>>()

		fun addMatterFunction(function: MatterManager.UpdateAction.Function, value: ImpreciseFraction): UpdateConfiguration {
			matterFunctions.add(MatterManager.UpdateAction.BoundFunction(function, value))
			matter = null
			return this
		}

		fun addComplexityFunction(function: MatterManager.UpdateAction.Function, value: Double): UpdateConfiguration {
			complexityFunctions.add(MatterManager.UpdateAction.BoundFunction(function, value))
			matter = null
			return this
		}

		fun addPriorityFunction(function: MatterManager.UpdateAction.Function, value: Int): UpdateConfiguration {
			priorityFunctions.add(MatterManager.UpdateAction.BoundFunction(function, value))
			matter = null
			return this
		}

		override fun build(): MatterManager.UpdateAction {
			check(!(matterFunctions.isNotEmpty() && matter != null)) { "Can't have both matter value and matter value functions be defined at the same time" }
			check(!(complexityFunctions.isNotEmpty() && complexity != null)) { "Can't have both complexity and complexity functions be defined at the same time" }
			check(!(priorityFunctions.isNotEmpty() && priority != null)) { "Can't have both priority and priority functions be defined at the same time" }

			check(
				matterFunctions.isNotEmpty() || matter != null ||
				complexityFunctions.isNotEmpty() || complexity != null ||
				priorityFunctions.isNotEmpty() || priority != null
			) { "Update configuration is completely empty" }

			checkMatterValues()
			checkKeyTag()

			if (key != null) {
				return MatterManager.UpdateAction(
					key = key ?: throw ConcurrentModificationException(),
					matter = matter,
					complexity = complexity,
					priority = priority,
					errorOnFailure = errorOnFailure,
					matterFunctions = matterFunctions,
					complexityFunctions = complexityFunctions,
					priorityFunctions = priorityFunctions
				)
			} else {
				return MatterManager.UpdateAction(
					tag = tag ?: throw ConcurrentModificationException(),
					matter = matter,
					complexity = complexity,
					priority = priority,
					errorOnFailure = errorOnFailure,
					matterFunctions = matterFunctions,
					complexityFunctions = complexityFunctions,
					priorityFunctions = priorityFunctions
				)
			}
		}
	}

	fun insert(name: ResourceLocation, configurator: InsertConfiguration.() -> Unit): MatterManager.InsertAction {
		check(!actions.containsKey(name)) { "Already has action with name $name" }
		val config = InsertConfiguration(name)
		configurator.invoke(config)
		val built = config.build()
		actions[name] = built
		return built
	}

	fun delete(name: ResourceLocation, configurator: DeleteConfiguration.() -> Unit): MatterManager.DeleteAction {
		check(!actions.containsKey(name)) { "Already has action with name $name" }
		val config = DeleteConfiguration(name)
		configurator.invoke(config)
		val built = config.build()
		actions[name] = built
		return built
	}

	fun update(name: ResourceLocation, configurator: UpdateConfiguration.() -> Unit): MatterManager.UpdateAction {
		check(!actions.containsKey(name)) { "Already has action with name $name" }
		val config = UpdateConfiguration(name)
		configurator.invoke(config)
		val built = config.build()
		actions[name] = built
		return built
	}

	protected fun updateLocation(input: ResourceLocation, prefix: String): ResourceLocation {
		if (namespace != null) {
			if (input.namespace == namespace) {
				return ResourceLocation(input.namespace, prefix + input.path)
			} else {
				return ResourceLocation(namespace, prefix + input.namespace + "/" + input.path)
			}
		} else {
			return ResourceLocation(input.namespace, prefix + input.path)
		}
	}

	fun insert(name: ItemLike, configurator: InsertConfiguration.() -> Unit): MatterManager.InsertAction {
		return insert(updateLocation(name.asItem().registryName ?: throw NullPointerException("${name.asItem()} does not have registry name!"), "item/")) {
			key = name.asItem().registryName ?: throw ConcurrentModificationException()
			configurator.invoke(this)
		}
	}

	fun delete(name: ItemLike, configurator: DeleteConfiguration.() -> Unit): MatterManager.DeleteAction {
		return delete(updateLocation(name.asItem().registryName ?: throw NullPointerException("${name.asItem()} does not have registry name!"), "item/")) {
			key = name.asItem().registryName ?: throw ConcurrentModificationException()
			configurator.invoke(this)
		}
	}

	fun update(name: ItemLike, configurator: UpdateConfiguration.() -> Unit): MatterManager.UpdateAction {
		return update(updateLocation(name.asItem().registryName ?: throw NullPointerException("${name.asItem()} does not have registry name!"), "item/")) {
			key = name.asItem().registryName ?: throw ConcurrentModificationException()
			configurator.invoke(this)
		}
	}

	fun insert(name: TagKey<Item>, configurator: InsertConfiguration.() -> Unit): MatterManager.InsertAction {
		return insert(updateLocation(name.location, "tag/")) {
			tag = name
			configurator.invoke(this)
		}
	}

	fun delete(name: TagKey<Item>, configurator: DeleteConfiguration.() -> Unit): MatterManager.DeleteAction {
		return delete(updateLocation(name.location, "tag/")) {
			tag = name
			configurator.invoke(this)
		}
	}

	fun update(name: TagKey<Item>, configurator: UpdateConfiguration.() -> Unit): MatterManager.UpdateAction {
		return update(updateLocation(name.location, "tag/")) {
			tag = name
			configurator.invoke(this)
		}
	}

	fun insert(name: ResourceLocation, 	configurator: Consumer<InsertConfiguration>) = insert(name, configurator::accept)
	fun delete(name: ResourceLocation, 	configurator: Consumer<DeleteConfiguration>) = delete(name, configurator::accept)
	fun update(name: ResourceLocation, 	configurator: Consumer<UpdateConfiguration>) = update(name, configurator::accept)
	fun insert(name: ItemLike, 			configurator: Consumer<InsertConfiguration>) = insert(name, configurator::accept)
	fun delete(name: ItemLike, 			configurator: Consumer<DeleteConfiguration>) = delete(name, configurator::accept)
	fun update(name: ItemLike, 			configurator: Consumer<UpdateConfiguration>) = update(name, configurator::accept)
	fun insert(name: TagKey<Item>, 		configurator: Consumer<InsertConfiguration>) = insert(name, configurator::accept)
	fun delete(name: TagKey<Item>, 		configurator: Consumer<DeleteConfiguration>) = delete(name, configurator::accept)
	fun update(name: TagKey<Item>, 		configurator: Consumer<UpdateConfiguration>) = update(name, configurator::accept)

	fun insert(name: ResourceLocation, 	matterValue: ImpreciseFraction, complexity: Double, configurator: InsertConfiguration.() -> Unit = {}) = insert(name) {
		this.matter = matterValue
		this.complexity = complexity
		configurator.invoke(this)
	}

	fun insert(name: ItemLike, 			matterValue: ImpreciseFraction, complexity: Double, configurator: InsertConfiguration.() -> Unit = {}) = insert(name) {
		this.matter = matterValue
		this.complexity = complexity
		configurator.invoke(this)
	}

	fun insert(name: TagKey<Item>, 		matterValue: ImpreciseFraction, complexity: Double, configurator: InsertConfiguration.() -> Unit = {}) = insert(name) {
		this.matter = matterValue
		this.complexity = complexity
		configurator.invoke(this)
	}

	fun insert(name: ResourceLocation, 	matterValue: ImpreciseFraction, complexity: Double, priority: Int, configurator: InsertConfiguration.() -> Unit = {}) = insert(name) {
		this.matter = matterValue
		this.complexity = complexity
		this.priority = priority
		configurator.invoke(this)
	}

	fun insert(name: ItemLike, 			matterValue: ImpreciseFraction, complexity: Double, priority: Int, configurator: InsertConfiguration.() -> Unit = {}) = insert(name) {
		this.matter = matterValue
		this.complexity = complexity
		this.priority = priority
		configurator.invoke(this)
	}

	fun insert(name: TagKey<Item>, 		matterValue: ImpreciseFraction, complexity: Double, priority: Int, configurator: InsertConfiguration.() -> Unit = {}) = insert(name) {
		this.matter = matterValue
		this.complexity = complexity
		this.priority = priority
		configurator.invoke(this)
	}

	@JvmOverloads
	fun insert(name: ResourceLocation, 	matterValue: ImpreciseFraction, complexity: Double, configurator: Consumer<InsertConfiguration> = Consumer {  }) = insert(name, matterValue, complexity, configurator::accept)
	@JvmOverloads
	fun insert(name: ItemLike, 			matterValue: ImpreciseFraction, complexity: Double, configurator: Consumer<InsertConfiguration> = Consumer {  }) = insert(name, matterValue, complexity, configurator::accept)
	@JvmOverloads
	fun insert(name: TagKey<Item>, 		matterValue: ImpreciseFraction, complexity: Double, configurator: Consumer<InsertConfiguration> = Consumer {  }) = insert(name, matterValue, complexity, configurator::accept)

	@JvmOverloads
	fun insert(name: ResourceLocation, 	matterValue: ImpreciseFraction, complexity: Double, priority: Int, configurator: Consumer<InsertConfiguration> = Consumer {  }) = insert(name, matterValue, complexity, priority, configurator::accept)
	@JvmOverloads
	fun insert(name: ItemLike, 			matterValue: ImpreciseFraction, complexity: Double, priority: Int, configurator: Consumer<InsertConfiguration> = Consumer {  }) = insert(name, matterValue, complexity, priority, configurator::accept)
	@JvmOverloads
	fun insert(name: TagKey<Item>, 		matterValue: ImpreciseFraction, complexity: Double, priority: Int, configurator: Consumer<InsertConfiguration> = Consumer {  }) = insert(name, matterValue, complexity, priority, configurator::accept)

	/**
	 * Override this if you prefer The Mundane Way™ instead of calling [insert], [delete] and [update] directly
	 *
	 * Called inside [run]
	 */
	protected open fun addActions() {}

	protected val added = ArrayList<MatterManager.AbstractRegistryAction>()
	val addedView: List<MatterManager.AbstractRegistryAction> = Collections.unmodifiableList(added)

	final override fun run(output: CachedOutput) {
		addActions()

		for ((key, value) in actions) {
			DataProvider.saveStable(output, value.toJson(), pathProvider.json(key))
			added.add(value)
		}
	}

	override fun getName(): String {
		return "Matter Data"
	}
}
